﻿Imports System.Net
Imports System.Net.NetworkInformation
Imports System.Text

Public Class NetHALRPC
    Friend Enum FuncId As Byte
        SCAN = 0
        PROG = 1
        GETSINGLE = 2
        XCHG = 3
        NAME = 4
        EXEC = 5
        PWM = 6
        STEPPER = 7
        ADC24 = 8
    End Enum

    Friend Structure ScanReplayModulePropsStr
        Dim IPAddr As String
        Dim FWVer As String
        Dim Name As String
        Dim MAC As String
        Dim ProgStatus As String
    End Structure
    Friend Structure PinProps
        Dim Index As Int32
        Dim Prop As Byte
    End Structure
    Friend Structure GetSinglePinReplayStr
        Dim PinName As String
        Dim Prop() As Byte
    End Structure
    Friend Structure ScanRemoteModuleRequestStr
        Dim IPAddr As String
        Dim Port As Int32
    End Structure
    Friend Structure ProgModuleReplayPropsStr
        Dim IPAddr As String
        Dim ProgStatus As String
    End Structure
    Friend Structure XchgModuleReplayStr
        Dim IPAddr As String
        Dim ProgStatus As String
    End Structure
    Friend Structure NameModulePropsStr
        Dim IPAddr As String
        Dim Name As String
    End Structure

    ' =========================================================================
    '  LIVIO - 05/02/2019 - START
    ' =========================================================================
    Friend Structure PwmPinData
        Dim PinId As Byte                  ' Pin index (0 to 28)
        Dim Frequency As Single
    End Structure

    Friend Structure Adc24ConfigData
        Dim Sps As UInt16
        Dim MultipleSetups As Boolean
        Dim FiltersSingleCycle As Boolean
        Dim Adc24PreFilterType As Byte
        Dim Adc24PostFilterType As Byte
        Dim Adc24ActiveChannels As UInt16
        Dim Adc24BiasedChannels As UInt16
        Dim Adc24DiffChannels As UInt16
        Dim Adc24PseudoChannels As UInt16
        Dim Adc24ChannelsGains As Byte()  ' (originally was 3 byte)
    End Structure

    Friend Structure StepperPinData
        Dim PinId As Byte                  ' Pin index (0 to 28)
        Dim StepperSpeedHz As UInt32
        Dim StepperAccHz_per_mS As UInt16
    End Structure

    Friend Structure DataReceived
        Dim IPAddr As String
        Dim ReceivedOK As Boolean
    End Structure


    ' =========================================================================
    '  LIVIO - 05/02/2019 - END
    ' =========================================================================


    Private WithEvents _NetCom As NetCom = New NetCom

    Friend Function ScanAdapters(ByVal LocalPort As Int32) As Int32
        Return _NetCom.ScanAdapters(LocalPort)
    End Function

    Friend Sub GetConnectionStats(ByVal SourceIP As String, ByRef RequestCounter As Int32, ByRef ReplayCounter As Int32)
        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(SourceIP, ipa) = True Then
            _NetCom.GetConnectionStats(ipa, RequestCounter, ReplayCounter)
        End If
    End Sub

    Friend Sub GetTotalConnectionsStats(ByRef RequestCounter As Int32, ByRef ReplayCounter As Int32)
        _NetCom.GetTotalConnectionsStats(RequestCounter, ReplayCounter)
    End Sub

    Friend Sub ResetConnectionStats()
        _NetCom.ResetConnectionsStats()
    End Sub

    Friend Sub CloseAdapters()
        _NetCom.CloseAdapters()
    End Sub

    Friend Sub ScanLocalModuleRequest()
        Dim data(12 - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev [uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.SCAN
        Index += 1

        _NetCom.FirstTXToLAN(data)
    End Sub

    Friend Sub ScanRemoteModuleRequest(ByVal ModuleProps As NetHALRPC.ScanRemoteModuleRequestStr)
        Dim data(12 - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev[uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.SCAN
        Index += 1

        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleProps.IPAddr, ipa) = True Then
            _NetCom.FirstTXToWAN(ipa, ModuleProps.Port, data)
        End If
    End Sub

    Friend Event ScanModuleReplay As ScanModuleReplayHandler
    Friend Delegate Sub ScanModuleReplayHandler(ByVal ModuleProps As NetHALRPC.ScanReplayModulePropsStr, ByVal PinsProps() As String)

    Friend Sub ProgAllModuleRequest(ByVal ModuleIP As String, ByVal PinsProps() As NetHALRPC.PinProps)
        Dim data(12 + 1 + PinsProps.Length - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev[uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.PROG
        Index += 1
        'TotalPinsNumber[unt8]
        data(Index) = CType(PinsProps.Length, Byte)
        Index += 1
        For Each PinProp As NetHALRPC.PinProps In PinsProps
            data(Index) = PinProp.Prop
            Index += 1
        Next

        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleIP, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event ProgModuleReplay As ProgAllModuleReplayHandler
    Friend Delegate Sub ProgAllModuleReplayHandler(ByVal ModuleProps As NetHALRPC.ProgModuleReplayPropsStr)

    Friend Sub GetSinglePinRequest(ByVal ModuleIP As String, ByVal PinIndex As UInt32)
        Dim data(12 + 1 - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev[uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.GETSINGLE
        Index += 1
        'PinIndex[unt8]
        data(Index) = CType(PinIndex, Byte)
        Index += 1

        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleIP, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event GetSinglePinReplay As GetSinglePinReplayHandler
    Friend Delegate Sub GetSinglePinReplayHandler(ByVal ModuleProps As String, ByVal PinsProps As NetHALRPC.GetSinglePinReplayStr)

    Friend Sub XchgSingleModuleRequest(ByVal ModuleIP As String, ByVal PinsOutValue() As Single)
        Dim data(12 + 1 + (PinsOutValue.Length * 4) - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev[uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.XCHG
        Index += 1
        'TotalValues[uint8]
        data(Index) = CType(PinsOutValue.Length, Byte)
        Index += 1
        For Each PinOutValue As Single In PinsOutValue
            Dim OutValueB() As Byte = BitConverter.GetBytes(PinOutValue)
            Array.Copy(OutValueB, 0, data, Index, 4)
            Index += 4
        Next

        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleIP, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event XchgSingleModuleReplay As XchgSingleModuleReplayHandler
    Friend Delegate Sub XchgSingleModuleReplayHandler(ByVal ModuleProps As NetHALRPC.XchgModuleReplayStr, ByVal PinsIn() As Single)

    Friend Sub NameModuleRequest(ByVal ModuleName As NetHALRPC.NameModulePropsStr)
        Dim data(12 + 1 + ModuleName.Name.Length - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev [uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.NAME
        Index += 1
        'ModuleNameSize[uint8]
        Dim ModuleNameSizeB As Byte = CByte(ModuleName.Name.Length)
        data(Index) = ModuleNameSizeB
        Index += 1
        'ModuleName[32 * uint8]
        Dim ModuleNameB() As Byte = System.Text.Encoding.Default.GetBytes(ModuleName.Name)
        Array.Copy(ModuleNameB, 0, data, Index, ModuleNameSizeB)
        Index += ModuleNameSizeB

        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleName.IPAddr, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event NameModuleReplay As NameModuleReplayHandler
    Friend Delegate Sub NameModuleReplayHandler(ByVal ModuleName As NetHALRPC.NameModulePropsStr)

    Friend Sub ExecModuleRequest(ByVal ModuleIP As String)
        Dim data(12 - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev [uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.EXEC
        Index += 1

        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleIP, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event FuotaModuleReplay As FuotaModuleReplayHandler
    Friend Delegate Sub FuotaModuleReplayHandler(ByVal ModuleName As NetHALRPC.NameModulePropsStr)

    Friend Sub PwmPinDataRequest(ByVal ModuleIP As String, ByVal PwmPinData As NetHALRPC.PwmPinData)
        Dim data(12 + 1 + 4 - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev[uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who[uint8]
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.PWM
        Index += 1
        'PinIndex[uint8]
        data(Index) = CType(PwmPinData.PinId, Byte)
        Index += 1
        'Freq[single]
        Dim OutValueB() As Byte = BitConverter.GetBytes(PwmPinData.Frequency)
        Array.Copy(OutValueB, 0, data, Index, 4)
        Index += 4

        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleIP, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event PwmPinDataReplay As PwmPinDataReplayHandler
    Friend Delegate Sub PwmPinDataReplayHandler(ByVal PwmPinDataReceived As NetHALRPC.DataReceived)

    ' =========================================================================
    Friend Sub StepperPinDataRequest(ByVal ModuleIP As String, ByVal StepperPinData As NetHALRPC.StepperPinData)
        Dim data(12 + 1 + 4 + 2 - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev[uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who([uint8])
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.STEPPER
        Index += 1
        ' -----------------------------------------------------------------------
        Dim b() As Byte
        b = BitConverter.GetBytes(StepperPinData.PinId)
        Array.Copy(b, 0, data, Index, 1)
        Index += 1
        b = BitConverter.GetBytes(StepperPinData.StepperSpeedHz)
        Array.Copy(b, 0, data, Index, 4)
        Index += 4
        b = BitConverter.GetBytes(StepperPinData.StepperAccHz_per_mS)
        Array.Copy(b, 0, data, Index, 2)
        Index += 2
        ' -----------------------------------------------------------------------
        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleIP, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event StepperPinDataReplay As StepperPinDataReplayHandler
    Friend Delegate Sub StepperPinDataReplayHandler(ByVal StepperPinDataReceived As NetHALRPC.DataReceived)

    ' =========================================================================
    Friend Sub Adc24ConfigRequest(ByVal ModuleIP As String, ByVal Adc24ConfigData As NetHALRPC.Adc24ConfigData)
        Dim data(12 + 17 - 1) As Byte
        'Protocol[6 * uint8]
        Dim Index As Int32 = 0
        Dim ProtocolB() As Byte = System.Text.Encoding.Default.GetBytes(_NetCom._ComProt)
        Array.Copy(ProtocolB, 0, data, Index, 6)
        Index += 6
        'ProtocolRev[uint16]
        Dim ProtocolVerB() As Byte = BitConverter.GetBytes(_NetCom._ComProtVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Who([uint8])
        data(Index) = Asc("H")
        Index += 1
        'FirmwareRev[uint16]
        Dim FirmwareVerB() As Byte = BitConverter.GetBytes(_NetCom._FirmwareVer)
        Array.Copy(ProtocolVerB, 0, data, Index, 2)
        Index += 2
        'Command[uint8]
        data(Index) = FuncId.ADC24
        Index += 1
        ' -----------------------------------------------------------------------
        Dim b() As Byte
        b = BitConverter.GetBytes(Adc24ConfigData.Sps)
        Array.Copy(b, 0, data, Index, 2)
        Index += 2
        data(Index) = CByte(Adc24ConfigData.MultipleSetups)
        Index += 1
        data(Index) = CByte(Adc24ConfigData.FiltersSingleCycle)
        Index += 1
        data(Index) = CByte(Adc24ConfigData.Adc24PreFilterType)
        Index += 1
        data(Index) = CByte(Adc24ConfigData.Adc24PostFilterType)
        Index += 1
        b = BitConverter.GetBytes(Adc24ConfigData.Adc24ActiveChannels)
        Array.Copy(b, 0, data, Index, 2)
        Index += 2
        b = BitConverter.GetBytes(Adc24ConfigData.Adc24BiasedChannels)
        Array.Copy(b, 0, data, Index, 2)
        Index += 2
        b = BitConverter.GetBytes(Adc24ConfigData.Adc24DiffChannels)
        Array.Copy(b, 0, data, Index, 2)
        Index += 2
        b = BitConverter.GetBytes(Adc24ConfigData.Adc24PseudoChannels)
        Array.Copy(b, 0, data, Index, 2)
        Index += 2
        data(Index) = CByte(Adc24ConfigData.Adc24ChannelsGains(0))
        Index += 1
        data(Index) = CByte(Adc24ConfigData.Adc24ChannelsGains(1))
        Index += 1
        data(Index) = CByte(Adc24ConfigData.Adc24ChannelsGains(2))
        Index += 1
        ' -----------------------------------------------------------------------
        Dim ipa As IPAddress = Nothing
        If IPAddress.TryParse(ModuleIP, ipa) = True Then
            _NetCom.TX(ipa, data)
        End If
    End Sub

    Friend Event Adc24ConfigDataReplay As Adc24ConfigDataReplayHandler
    Friend Delegate Sub Adc24ConfigDataReplayHandler(ByVal Adc24ConfigDataReceived As NetHALRPC.DataReceived)


    Private Sub RXEventHandler(ByVal IP As IPAddress, ByVal Data() As Byte) Handles _NetCom.RXEvent
        'Protocol[6]
        Dim Index As Int32 = 0
        Dim ProtocolB(6 - 1) As Byte
        Array.Copy(Data, Index, ProtocolB, 0, 6)
        Index += 6
        If Encoding.Default.GetString(ProtocolB) = _NetCom._ComProt Then
            'ProtocolVer[6 * uint16]
            Dim ProtocolVer As Int16 = BitConverter.ToInt16(Data, Index)
            Index += 2
            'Who[uint8]
            Dim Who As Byte = Data(Index)
            Index += 1
            If Who = CByte(Asc("M")) Then
                'FirmwareVer[uint16]
                Dim FirmwareVer As Int16 = BitConverter.ToInt16(Data, Index)
                Index += 2
                'Command[uint8]
                Dim Command As FuncId = CType(Data(Index), FuncId)
                Index += 1
                Select Case Command
                    Case FuncId.SCAN
                        Dim ModuleProps As NetHALRPC.ScanReplayModulePropsStr
                        ModuleProps.IPAddr = IP.ToString
                        ModuleProps.FWVer = FirmwareVer.ToString
                        'MAC[6 * uint8]
                        Dim MACB(CByte(6 - 1)) As Byte
                        Array.Copy(Data, Index, MACB, 0, 6)
                        ModuleProps.MAC = MACB(0).ToString("X2") + MACB(1).ToString("X2") + MACB(2).ToString("X2") + MACB(3).ToString("X2") + MACB(4).ToString("X2") + MACB(5).ToString("X2")
                        Index += 6
                        'ModuleNameSize[uint8]
                        Dim ModuleNameSizeB As Byte = Data(Index)
                        Index += 1
                        'ModuleName[32 * uint8]
                        Dim ModuleNameB(ModuleNameSizeB - 1) As Byte
                        Array.Copy(Data, Index, ModuleNameB, 0, ModuleNameSizeB)
                        Index += 32
                        ModuleProps.Name = Encoding.Default.GetString(ModuleNameB)
                        'PinoutType[uint8]
                        Dim PinoutTypeB As Byte = Data(Index)
                        Index += 1
                        ModuleProps.FWVer = ModuleProps.FWVer + "-" + CStr(PinoutTypeB)
                        'ProgStatus[uint8]
                        Dim ProgStatusB As Byte = Data(Index)
                        Index += 1
                        ModuleProps.ProgStatus = ChrW(ProgStatusB)
                        'TotalPinsNumber[uint8]
                        Dim TotalPinsNumberB As Byte = Data(Index)
                        Index += 1
                        Dim PinsProps(TotalPinsNumberB - 1) As String
                        If TotalPinsNumberB > 0 Then
                            For i As Byte = 0 To TotalPinsNumberB - CByte(1)
                                'PinName[uint8]
                                Dim PinNameB As Byte
                                PinNameB = Data(Index)
                                Index += 1
                                PinsProps(i) = CStr(PinNameB)
                            Next
                        End If
                        RaiseEvent ScanModuleReplay(ModuleProps, PinsProps)
                    Case FuncId.PROG
                        Dim ModuleProps As NetHALRPC.ProgModuleReplayPropsStr
                        ModuleProps.IPAddr = IP.ToString
                        'ProgStatus[uint8]
                        Dim ProgStatusB As Byte = Data(Index)
                        Index += 1
                        ModuleProps.ProgStatus = ChrW(ProgStatusB)
                        RaiseEvent ProgModuleReplay(ModuleProps)
                    Case FuncId.XCHG
                        Dim ModuleProps As NetHALRPC.XchgModuleReplayStr
                        ModuleProps.IPAddr = IP.ToString
                        'ProgStatus[uint8]
                        Dim ProgStatusB As Byte = Data(Index)
                        Index += 1
                        ModuleProps.ProgStatus = ChrW(ProgStatusB)
                        'TotalValues[uint8]
                        Dim TotalValuesB As Byte = Data(Index)
                        Index += 1
                        Dim PinsValues(TotalValuesB - 1) As Single
                        If (TotalValuesB > 0) Then
                            For i As Byte = 0 To TotalValuesB - CByte(1)
                                'Value[single]
                                PinsValues(i) = BitConverter.ToSingle(Data, Index)
                                Index += 4
                            Next
                        End If

                        ' ''TODO - Adc16 input = 0 --------------------

                        RaiseEvent XchgSingleModuleReplay(ModuleProps, PinsValues)


                    Case FuncId.GETSINGLE
                        Dim ModuleIP As String
                        Dim SinglePin As NetHALRPC.GetSinglePinReplayStr = New NetHALRPC.GetSinglePinReplayStr
                        ModuleIP = IP.ToString
                        'PinName[uint8]
                        Dim PinNameB As Byte
                        PinNameB = Data(Index)
                        Index += 1
                        SinglePin.PinName = CStr(PinNameB)
                        'PinTypesSize[uint8]
                        Dim PinTypesSizeB As Byte = Data(Index)
                        Index += 1
                        If (PinTypesSizeB > 0) Then
                            'PinTypes [PinTypesSizeB * uint8]
                            Dim PinTypesB(PinTypesSizeB - 1) As Byte
                            Array.Copy(Data, Index, PinTypesB, 0, PinTypesSizeB)
                            Index += PinTypesSizeB
                            SinglePin.Prop = PinTypesB
                        End If
                        RaiseEvent GetSinglePinReplay(ModuleIP, SinglePin)
                    Case FuncId.NAME
                        Dim ModuleProps As NetHALRPC.NameModulePropsStr
                        ModuleProps.IPAddr = IP.ToString
                        'ModuleNameSize[uint8]
                        Dim ModuleNameSizeB As Byte = Data(Index)
                        Index += 1
                        'ModuleName[ModuleNameSize * uint8]
                        Dim ModuleNameB(ModuleNameSizeB - 1) As Byte
                        Array.Copy(Data, Index, ModuleNameB, 0, ModuleNameSizeB)
                        Index += ModuleNameSizeB
                        ModuleProps.Name = Encoding.Default.GetString(ModuleNameB)
                        RaiseEvent NameModuleReplay(ModuleProps)
                    Case FuncId.PWM
                        Dim PwmPinDataReceived As NetHALRPC.DataReceived
                        PwmPinDataReceived.IPAddr = IP.ToString
                        'ReceivedOK[uint8]
                        Dim ProgStatusB As Byte = Data(Index)
                        Index += 1
                        PwmPinDataReceived.ReceivedOK = CBool(ProgStatusB)
                        RaiseEvent PwmPinDataReplay(PwmPinDataReceived)

                    Case FuncId.STEPPER
                        Dim StepperPinDataReceived As NetHALRPC.DataReceived
                        StepperPinDataReceived.IPAddr = IP.ToString
                        'ReceivedOK[uint8]
                        Dim ProgStatusB As Byte = Data(Index)
                        Index += 1
                        StepperPinDataReceived.ReceivedOK = CBool(ProgStatusB)
                        RaiseEvent StepperPinDataReplay(StepperPinDataReceived)

                    Case FuncId.ADC24
                        Dim Adc24ConfigDataReceived As NetHALRPC.DataReceived
                        Adc24ConfigDataReceived.IPAddr = IP.ToString
                        'ReceivedOK[uint8]
                        Dim ProgStatusB As Byte = Data(Index)
                        Index += 1
                        Adc24ConfigDataReceived.ReceivedOK = CBool(ProgStatusB)
                        RaiseEvent Adc24ConfigDataReplay(Adc24ConfigDataReceived)
                End Select
            End If
        End If
    End Sub

End Class
