﻿
Public Class Form1

    Private i32 As Int32 = 33
    Private ui16 As UInt16 = CUShort(i32)

    Private EventsLock As Object = New Object

    Private Sub Form1_Load(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Load
        SchedulerMaxPrecision()
        Load_INI()
        Load_ConfigDatabase()
        LoadSlotNames()
        LoadCommOptions()
        SetLocales()
        ToolTips_Init()
        ToolStrip1.Renderer = New ToolStripButtonRenderer
        ' ---------------------------------------------------------------- REAL-TIME
        System.Threading.Thread.CurrentThread.Priority = System.Threading.ThreadPriority.Normal
        System.Diagnostics.Process.GetCurrentProcess().PriorityClass = ProcessPriorityClass.RealTime
        ' ---------------------------------------------------------------- Init and recognize MODULES
        CommandSlot_StartRecognizeMarker()
        TheSystem_InitModules()
        TheSystem_AssignConfigurationToModules()
        TheSystem_SetParkingPositions()
        TheSystem_StartTimers()
        CommandSlot_WriteModulesCount()
        EnsureSafeValueToTextSlots()
        ' ---------------------------------------------------------------- Timer1 = 100mS = 10Hz 
        Timer_10Hz.Interval = 100
        Timer_10Hz.Start()
        ' ---------------------------------------------------------------- Timer_60Hz = 16.6mS = 60Hz
        Timer_60Hz.Interval = 10
        Timer_60Hz.Start()
        ' ---------------------------------------------------------------- WINDOW
        If Not OperatingSystemIsWindows Then
            Me.MinimumSize = New Size(650, 500)
        End If
        If Not ConfigurationIsValid Then
            WindowState = FormWindowState.Normal
        End If
        If Form2_VisibleAtStart Then
            Form2.Show()
        Else
            Form2.Visible = False
        End If
        Text = AppTitleAndVersion()
        ' ----------------------------------------------------------------
        EventsAreEnabled = True
        ' ----------------------------------------------------------------
        ListView_SelectLine(MyListView1, 0)
        ShowProps()
        Form2.RestoreSelectedPins()

        EventsAreEnabled = False
        MyListView1.Focus()
        ShowInTaskbar = False
        ShowInTaskbar = True
        EventsAreEnabled = True

        Refresh()
        Form2.Refresh()
        Me.Opacity = 1
        Form2.Opacity = 1

        InvalidConfigMessage()
    End Sub

    Private Sub Form1_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        If Not EventsAreEnabled Then Return
        CloseAllAndExit()
    End Sub

    Private Sub Form1_LocationChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.LocationChanged
        If Not EventsAreEnabled Then Return
        LimitFormPosition(Me)
    End Sub

    Private Sub CloseAllAndExit()
        TheSystem_StopTimers()
        TheSystem_SetParkingPositions()
        TheSystem_UpdateAndSaveConfigDatabase()
        Save_INI()
        SchedulerDefaultPrecision()
    End Sub

    Private Sub InvalidConfigMessage()
        If ConfigurationIsValid = False Then
            If DuplicateNames Then
                MessageBox.Show(Msg_Warning + vbCrLf + vbCrLf + _
                                Msg_Duplicate1 + vbCrLf + vbCrLf + _
                                Msg_Duplicate2, _
                                Msg_HalMessage, _
                                MessageBoxButtons.OK)
            Else
                MessageBox.Show(Msg_Warning + vbCrLf + vbCrLf + _
                    Msg_Validate2 + vbCrLf + vbCrLf + _
                    Msg_Validate3 + vbCrLf + vbCrLf + _
                    Msg_Validate4 + vbCrLf + vbCrLf + _
                    Msg_Validate5, _
                    Msg_HalMessage, _
                    MessageBoxButtons.OK)
            End If
        End If
    End Sub


    ' ===================================================================================
    '  MenuStrip and ToolStrip accepting the first click
    '  If the form receives a WM_PARENTNOTIFY (528) message and is not focused 
    '  then the form is activated before to exec the message
    ' ===================================================================================
    <DebuggerStepThrough()> _
    Protected Overrides Sub WndProc(ByRef m As Message)
        If m.Msg = 528 AndAlso Not Me.Focused Then
            Me.Activate()
        End If
        MyBase.WndProc(m)
    End Sub

    ' ===================================================================================
    '   MenuStrip and ToolStrip Gradients
    ' ===================================================================================
    Private Sub MenuStrip1_Paint(ByVal sender As Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles MenuStrip1.Paint
        Dim bounds As New Rectangle(0, 0, _
                                    MenuStrip1.Width, MenuStrip1.Height)
        Dim brush As New Drawing2D.LinearGradientBrush(bounds, _
                                                       Color.FromArgb(240, 240, 240), _
                                                       Color.FromArgb(200, 200, 200), _
                                                       Drawing2D.LinearGradientMode.Horizontal)
        e.Graphics.FillRectangle(brush, bounds)
    End Sub
    Private Sub ToolStrip1_Paint(ByVal sender As System.Object, ByVal e As System.Windows.Forms.PaintEventArgs) Handles ToolStrip1.Paint
        Dim bounds As New Rectangle(0, 0, _
                                    ToolStrip1.Width, ToolStrip1.Height)
        Dim brush As New Drawing2D.LinearGradientBrush(bounds, _
                                                       Color.White, _
                                                       Color.FromArgb(200, 200, 200), _
                                                       Drawing2D.LinearGradientMode.Vertical)
        e.Graphics.FillRectangle(brush, bounds)
    End Sub

    ' ===================================================================================
    '   ToolStrip PressedButton color
    ' ===================================================================================
    Class ToolStripButtonRenderer
        Inherits System.Windows.Forms.ToolStripProfessionalRenderer

        Protected Overrides Sub OnRenderButtonBackground(ByVal e As ToolStripItemRenderEventArgs)
            Dim btn As ToolStripButton = CType(e.Item, ToolStripButton)
            If btn IsNot Nothing AndAlso btn.CheckOnClick AndAlso btn.Checked Then
                Dim bounds As Rectangle = New Rectangle(0, 0, e.Item.Width - 1, e.Item.Height - 1)
                Dim brush As New Drawing2D.LinearGradientBrush(bounds, _
                                                               Color.Gold, _
                                                               Color.FromArgb(250, 250, 250), _
                                                               Drawing2D.LinearGradientMode.Vertical)
                e.Graphics.FillRectangle(brush, bounds)
                e.Graphics.DrawRectangle(Pens.Orange, bounds)
            Else
                MyBase.OnRenderButtonBackground(e)
            End If
        End Sub
    End Class


    ' =======================================================================================
    '   MENU FILE
    ' =======================================================================================
    Private Sub Menu_File_OpenProgramFolder_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_OpenProgramFolder.Click
        Process.Start(Application.StartupPath)
    End Sub
    Private Sub Menu_File_EditSlotNames_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_EditSlotNames.Click
        Me.Opacity = 0.7
        If Not FileExists(SlotFilePathAndName) Then IO.File.Create(SlotFilePathAndName)
        ProcessStartAndWait(SlotFilePathAndName)
        Me.Opacity = 1
    End Sub
    Private Sub Menu_File_EditConfigurations_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_EditConfigurations.Click
        Me.Opacity = 0.7
        ProcessStartAndWait(PlatformAdjustedFileName(Application.StartupPath & "\" & AssemblyName() & "_ConfigDatabase.txt"))
        Cmd_Recognize()
        Me.Opacity = 1
    End Sub
    Private Sub Menu_File_EditCommOptions_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_EditCommOptions.Click
        Me.Opacity = 0.7
        Dim fname As String = Application.StartupPath & "\CommOptions.txt"
        If Not FileExists(fname) Then IO.File.Create(fname)
        ProcessStartAndWait(PlatformAdjustedFileName(fname))
        Me.Opacity = 1
        LoadCommOptions()
        Cmd_Recognize()
    End Sub
    Private BackupFolder As String = My.Computer.FileSystem.SpecialDirectories.MyDocuments + "\Theremino\IotHAL_Backups\"
    Private HalFolder As String = Application.StartupPath + "\"
    Private Sub Menu_File_BackupConfigurations_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_BackupConfigurations.Click
        Dim DestFolder As String = Date.Now.ToString("yyyy_MM_dd_HH_mm_ss")
        DestFolder = InputBox(vbCrLf + "Backup folder name?", "Theremino IotHAL - Save to backup folder", DestFolder)
        If DestFolder <> "" Then
            DestFolder = BackupFolder + DestFolder + "\"
            Try
                My.Computer.FileSystem.CreateDirectory(DestFolder)
                CopyFileIfExists(HalFolder, DestFolder, AssemblyName() & "_ConfigDatabase.txt")
                CopyFileIfExists(HalFolder, DestFolder, "SlotNames.txt")
                CopyFileIfExists(HalFolder, DestFolder, "Theremino_IotHAL_INI.txt")
                CopyFileIfExists(HalFolder, DestFolder, "Theremino_SlotViewer_INI.txt")
            Catch
            End Try
        End If
    End Sub
    Private Sub Menu_File_LoadConfigurations_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_LoadConfigurations.Click
        Dim SourceFolder As String = SelectFolder(BackupFolder)
        If SourceFolder <> "" Then
            CopyFileIfExists(SourceFolder, HalFolder, AssemblyName() & "_ConfigDatabase.txt")
            CopyFileIfExists(SourceFolder, HalFolder, "SlotNames.txt")
            CopyFileIfExists(SourceFolder, HalFolder, "Theremino_IotHAL_INI.txt")
            CopyFileIfExists(SourceFolder, HalFolder, "Theremino_SlotViewer_INI.txt")
            TheSystem_AssignConfigurationToModules()
        End If
    End Sub

    Friend Sub CopyFileIfExists(ByVal srcFolder As String, ByVal destFolder As String, ByVal fileName As String)
        Dim src As String = srcFolder + fileName
        Dim dest As String = destFolder + fileName
        If IO.File.Exists(src) Then
            IO.File.Copy(src, dest, True)
        End If
    End Sub
    Friend Function SelectFolder(ByVal folder As String) As String
        Dim FBD As System.Windows.Forms.FolderBrowserDialog
        FBD = New System.Windows.Forms.FolderBrowserDialog()
        With FBD
            .Reset()
            .ShowNewFolderButton = False
            ' -- USE ROOTFOLDER TO LIMIT USER CHOOSE AREA -- ( No RootFolder NO limits )
            '.RootFolder = Environment.SpecialFolder.MyDocuments
            .RootFolder = Environment.SpecialFolder.MyComputer
            ' --------------------------------------------------------------------------
            .SelectedPath = folder
            .Description = vbCr & "Select the backup folder"
            SendKeys.Send("{TAB}{TAB}{RIGHT}")
            If .ShowDialog = DialogResult.OK Then
                Return .SelectedPath + "\"
            End If
        End With
        Return ""
    End Function

    Private Sub Menu_File_Exit_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_File_Exit.Click
        CloseAllAndExit()
        Me.Close()
    End Sub

    ' =======================================================================================
    '   MENU TOOLS
    ' =======================================================================================
    Private Sub PrepareSelectedModuleForWiFiProgramming_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Tools_PrepareModule.Click
        Dim m As Module_InOut = FindModuleByListLine(SelectedLine)
        If m IsNot Nothing Then
            _NetHALRPC.ExecModuleRequest(m.IpAddr)
        End If
    End Sub
    Private Sub ReinitSelectedModuleSlots_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Tools_ReinitSlots.Click
        Dim m As Module_InOut = FindModuleByListLine(SelectedLine)
        If m IsNot Nothing Then
            Dim n As Int32 = CInt(Val(InputBox(vbCrLf + "First Slot ?", "IotHAL - Reinit Slots for module " + m.GetName)))
            If n > 0 Then
                m.InitSlots(n)
                ThereminoSystem.TheSystem_ListComponents()
                ListView_SetAllLineColors()
                TestSlotConflicts()
                TheSystem_UpdateAndSaveConfigDatabase()
            End If
        End If
    End Sub

    ' =======================================================================================
    '   MENU LANGUAGE
    ' =======================================================================================
    Private Sub Menu_Language_DropDownOpening(ByVal sender As Object, ByVal e As System.EventArgs) Handles Menu_Language.DropDownOpening
        For Each item As ToolStripMenuItem In Menu_Language.DropDownItems
            If item.Name.EndsWith(Language, StringComparison.InvariantCultureIgnoreCase) Then
                item.Select()
            End If
        Next
    End Sub
    Private Sub Menu_Language_Deutsch_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Menu_Language_Deutsch.Click
        Language = "Deutsch"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub
    Private Sub Menu_Language_English_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Menu_Language_English.Click
        Language = "English"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub
    Private Sub Menu_Language_Espanol_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Menu_Language_Espanol.Click
        Language = "Espanol"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub
    Private Sub Menu_Language_Portoguese_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Menu_Language_Portoguese.Click
        Language = "Portugues"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub
    Private Sub Menu_Language_Francais_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Menu_Language_Francais.Click
        Language = "Francais"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub
    Private Sub Menu_Language_Italian_Click(ByVal sender As Object, ByVal e As System.EventArgs) Handles Menu_Language_Italian.Click
        Language = "Italian"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub
    Private Sub Menu_Language_Japanese_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Language_Japanese.Click
        Language = "Japanese"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub
    Private Sub Menu_Language_Chinese_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Language_Chinese.Click
        Language = "Chinese"
        SetLocales()
        Save_INI()
        ToolTips_Init()
        ShowProps()
    End Sub

    ' =======================================================================================
    '   MENU HELP
    ' =======================================================================================
    Private Sub Menu_Help_ProgramHelp_ENG_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_Help_ProgramHelp.Click
        OpenLocalizedHelp("Theremino_IotHAL_Help", ".pdf")
    End Sub
    Private Sub OpenLocalizedHelp(ByVal name As String, Optional ByVal ext As String = ".rtf")
        Dim LangName As String = Strings.Left(Language, 3).ToUpper.Replace("JAP", "JPN")
        Dim fname As String = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & "_" & LangName & ext)
        If FileExists(fname) Then
            Process.Start(fname)
        Else
            fname = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & "_ENG" & ext)
            If FileExists(fname) Then
                Process.Start(fname)
            Else
                fname = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & "_ITA" & ext)
                If FileExists(fname) Then
                    Process.Start(fname)
                Else
                    fname = PlatformAdjustedFileName(Application.StartupPath & "\Docs\" & name & ext)
                    If FileExists(fname) Then
                        Process.Start(fname)
                    End If
                End If
            End If
        End If
    End Sub

    ' =======================================================================================
    '   MENU ABOUT
    ' =======================================================================================
    Private Sub Menu_About_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Menu_About.Click
        Form_About.ShowDialog()
    End Sub

    ' =======================================================================================
    '   TOOLBAR
    ' =======================================================================================
    Private Sub ToolStripButton_Recognize_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_Recognize.Click
        Cmd_Recognize()
        ToolStripButton_Recognize.Checked = False
    End Sub
    Private Sub ToolStripButton_Validate_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_Validate.Click
        Cmd_Validate()
        ToolStripButton_Validate.Checked = False
    End Sub
    Private Sub ToolStripButton_EditCommOptions_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_CommOptions.Click
        Menu_File_EditCommOptions_Click(Nothing, Nothing)
        ToolStripButton_CommOptions.Checked = False
    End Sub
    Private Sub ToolStripButton_Disconnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles ToolStripButton_Disconnect.Click
        Cmd_DisconnectSelectedModule()
        ToolStripButton_Disconnect.Checked = False
    End Sub

    'Private Sub ToolStripButton_EditConfigurations_Click(ByVal sender As System.Object, ByVal e As System.EventArgs)
    '    Menu_File_EditConfigurations_Click(Nothing, Nothing)
    'End Sub
    'Private Sub ToolStripButton_EditSlotNames_Click(ByVal sender As System.Object, ByVal e As System.EventArgs)
    '    Menu_File_EditSlotNames_Click(Nothing, Nothing)
    'End Sub


    ' ===========================================================================================================
    '   Commands execution
    ' ===========================================================================================================
    Friend Sub Cmd_Recognize()
        CommandSlot_StartRecognizeMarker()
        EventsAreEnabled = False
        Dim OldSelectedLine As Int32 = SelectedLine
        Load_ConfigDatabase()
        Timer_10Hz.Stop()
        TheSystem_StopTimers()
        TheSystem_InitModules()
        TheSystem_AssignConfigurationToModules()
        GroupBox_ModuleProps.BackColor = Color.LightGoldenrodYellow
        ShowProps()
        TheSystem_StartTimers()
        Timer_10Hz.Start()
        EventsAreEnabled = True
        If OldSelectedLine < 0 Then OldSelectedLine = 0
        ListView_SelectLine(MyListView1, OldSelectedLine)
        Form2.RestoreSelectedPins()
        ' ------------------------------------------ eventually show the message
        InvalidConfigMessage()
        ' ------------------------------------------ write modules count
        CommandSlot_WriteModulesCount()
    End Sub

    Private Sub CommandSlot_StartRecognizeMarker()
        Slots.WriteSlot(CommandSlot, -1)
    End Sub
    Private Sub CommandSlot_WriteModulesCount()
        Slots.WriteSlot(CommandSlot, InOutModules.Count)
    End Sub

    Friend Sub Cmd_Validate()
        If Not EventsAreEnabled Then Return
        EventsAreEnabled = False
        Dim OldSelectedLine As Int32 = SelectedLine
        TheSystem_ValidateConfiguration()
        TheSystem_UpdateAndSaveConfigDatabase()
        TheSystem_StartTimers()
        EventsAreEnabled = True
        ListView_SelectLine(MyListView1, OldSelectedLine)
    End Sub
    Friend Sub Cmd_DisconnectSelectedModule()
        EventsAreEnabled = False
        Load_ConfigDatabase()
        Timer_10Hz.Stop()
        TheSystem_StopTimers()
        ' -----------------------------------------------------
        TheSystem_DisconnectModule(FindModuleByListLine(SelectedLine))
        TheSystem_ListComponents()
        ListView_SetAllLineColors()
        SelectedLine = -1
        ' -----------------------------------------------------
        ShowProps()
        TheSystem_StartTimers()
        Timer_10Hz.Start()
        EventsAreEnabled = True
        ListView_SelectLine(MyListView1, 0)
    End Sub


    ' ===========================================================================================================
    '   UPDATES
    ' ===========================================================================================================
    Private Sub Timer_10Hz_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Timer_10Hz.Tick
        If Not EventsAreEnabled Then Return
        If Not ConfigurationIsValid Then Return
        ' ----------------------------------------------------------------- Test Module Error
        TheSystem_TestModuleError()
        ' ----------------------------------------------------------------- Update SlotNames if file is changed
        TestSlotFile()
        ' ----------------------------------------------------------------- updates
        SyncLock EventsLock
            '
            If WindowState <> FormWindowState.Minimized Then
                ' ---------------------------------------------------------- UpdateListedValues
                TheSystem_UpdateListedValues()
                ' ---------------------------------------------------------- update RepeatFrequency
                Dim m As Module_InOut = FindModuleByListLine(SelectedLine)
                If m IsNot Nothing Then
                    lbl_RepeatFrequency.Text = m.CommFps.ToString("0")
                    If m.CommFps < 50 Then
                        lbl_RepeatFrequency.BackColor = Color.LightSalmon
                    Else
                        lbl_RepeatFrequency.BackColor = Color.Transparent
                    End If
                    lbl_ErrorRate.Text = m.ErrorRate.ToString("0.00", GCI)
                    If m.ErrorRate >= 0.1 Then
                        lbl_ErrorRate.BackColor = Color.LightSalmon
                        'm.SetSpeed()
                    Else
                        lbl_ErrorRate.BackColor = Color.Transparent
                    End If
                    If m.ErrorRate >= 0.01 Then
                        If ToolStripButton_BeepOnErrors.Checked Then Beep_RepetitionLimited(150)
                    End If
                End If
            End If
            ' ---------------------------------------------------------- edit
            If EditValue_MouseEditFlag Then
                If LeftMousePressed() Then
                    EditValue_MouseMove()
                Else
                    ShowCursor()
                    EditValue_MouseEditFlag = False
                End If
            End If
            '
        End SyncLock
    End Sub


    ' ===========================================================================================================
    '   Command Slot
    ' ===========================================================================================================
    Private CommandStatus As Int32 = 0
    Private Sub Timer_60Hz_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles Timer_60Hz.Tick
        '
        ' ----------------------------------------------------------------- Poll the TEXT COMMANDS SLOT
        PollTextCommandsSlot()
        '
        ' ----------------------------------------------------------------- Poll the NUMERIC COMMANDS SLOT
        PollNumericCommandsSlot()
    End Sub


    ' ===========================================================================================================
    '   EDIT VALUES
    ' ===========================================================================================================
    Private EditValue_MouseEditFlag As Boolean = False
    Private EditValue_Slot As Int32
    Private EditValue_Multiplier As Int32
    Private EditValue_StartCursorPos As Point
    Private EditValue_MaxValue As Single
    Private EditValue_MinValue As Single
    Private EditValue_IsServoPin As Boolean

    Private Sub MyListView1_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyListView1.MouseDown
        If Not ConfigurationIsValid Then Return
        StartMouseEdit()
    End Sub

    Private Sub StartMouseEdit()
        If EditValue_MouseEditFlag Then Return
        Dim column As Int32 = ListView_GetColumnIndex(MyListView1)
        Dim line As Int32 = ListView_GetLineIndex(MyListView1)
        If line >= 0 Then
            Dim slotString As String = MyListView1.Items(line).SubItems(3).Text
            If slotString <> "" Then
                Dim p As Pin = FindPinByListLine(line)
                If p IsNot Nothing AndAlso p.Direction <> Pin.Directions.Unused Then
                    HideCursor()
                    EditValue_MouseEditFlag = True
                    EditValue_StartCursorPos = Cursor.Position
                    ' ----------------------------------------------------------------------
                    EditValue_Slot = p.Slot
                    EditValue_IsServoPin = False
                    If p.GetPinType() = Pin.PinTypes.SERVO Then EditValue_IsServoPin = True
                    If p.GetPinType() = Pin.PinTypes.STEPPER Then
                        EditValue_MaxValue = Single.MaxValue
                        EditValue_MinValue = Single.MinValue
                    Else
                        EditValue_MaxValue = CSng(Math.Max(p.Value_Max, p.Value_Min))
                        EditValue_MinValue = CSng(Math.Min(p.Value_Max, p.Value_Min))
                    End If
                    EditValue_Multiplier = 1
                    Dim range As Int32 = CInt(Math.Abs(p.Value_Max - p.Value_Min))
                    If range > 500 Then EditValue_Multiplier = 2
                    If range > 1000 Then EditValue_Multiplier = 5
                    If range > 2000 Then EditValue_Multiplier = 10
                    If range > 5000 Then EditValue_Multiplier = 20
                    If range > 10000 Then EditValue_Multiplier = 50
                End If
            End If
        End If
    End Sub

    Private Sub MyListView1_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyListView1.MouseMove
        If Not EventsAreEnabled Then Return
        If EditValue_MouseEditFlag Then
            EditValue_MouseMove()
        End If
    End Sub

    Private Sub EditValue_MouseMove()
        Dim delta As Int32 = EditValue_StartCursorPos.Y - Cursor.Position.Y
        Cursor.Position = EditValue_StartCursorPos
        ChangeValueByDelta(delta)
    End Sub

    Private Sub ChangeValueByDelta(ByVal delta As Single)
        ' Do not limit the value if not moving
        If delta = 0 Then Return

        'Form2.StopTimer()
        EventsAreEnabled = False

        Dim v As Single
        v = Slots.ReadSlot(EditValue_Slot)

        If Single.IsNaN(v) Then
            v = EditValue_MinValue - 9
        End If
        '
        v += delta * EditValue_Multiplier
        v = Math.Min(v, EditValue_MaxValue)
        'v = Math.Max(v, EditValue_MinValue)
        '
        If v < EditValue_MinValue Then
            If v < -8 AndAlso EditValue_IsServoPin Then
                v = NAN_Sleep
            Else
                v = EditValue_MinValue
            End If
        End If
        '
        Slots.WriteSlot(EditValue_Slot, v)
        '
        Application.DoEvents()

        EventsAreEnabled = True
        'Form2.StartTimer()
    End Sub

    Private Sub MyListView1_MouseDoubleClick(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyListView1.MouseDoubleClick
        Dim p As Pin = GetSelectedPin()
        If p Is Nothing Then Return
        If p.Direction = Pin.Directions.Unused Then Return
        Form2.Show()
    End Sub

    Private Sub MyListView1_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles MyListView1.MouseUp
        Dim p As Pin = GetSelectedPin()
        If p Is Nothing Then Return
        If p.Direction = Pin.Directions.Unused Then Return
        Form2.SetPinDetails()
    End Sub

    ' ------------------------------------------------------------------
    '  Call ListViewResize from Form.Resize (not from ListView.resize)
    '  (otherwise the listview scroll produces problems)
    ' ------------------------------------------------------------------
    Private Sub Form1_Resize(ByVal sender As Object, ByVal e As System.EventArgs) Handles Me.Resize
        If Not EventsAreEnabled Then Return
        TheSystem_ListViewResize()
    End Sub

    ' ---------------------------------------------------------------------------------------------------
    '  MyListView1.Activation = ItemActivation.Standard   <<<< USE THIS OR THE CURSOR CAN NOT BE CHANGED
    ' ---------------------------------------------------------------------------------------------------
    Private BlankCursor As Cursor = New Cursor(New IO.MemoryStream(My.Resources.Blank))
    Private Sub HideCursor()
        Cursor.Hide()
        MyListView1.Cursor = BlankCursor
    End Sub
    Private Sub ShowCursor()
        Cursor.Show()
        MyListView1.Cursor = Cursors.Hand
    End Sub



    ' ===========================================================================================================
    '   SET MODULE AND PIN PROPERTIES
    ' ===========================================================================================================
    Private Sub btn_ModuleName_ClickButtonArea(ByVal Sender As System.Object, ByVal e As System.EventArgs) Handles btn_ModuleName.ClickButtonArea
        Dim newName As String
        newName = InputBox("New name for the selected module", _
                           "Edit Module name", _
                           cmb_ModuleNames.Text, _
                           Me.Right - 370, _
                           Me.Top + 125)

        newName = newName.Trim
        If newName <> "" Then
            ' ----------------------------------------- if the new name already exists
            If FindConfigByName(newName) >= 0 Then
                MessageBox.Show("The name """ + newName + """ already exists." + vbCrLf + _
                                "the existing configuration will be used.")
            End If
            Change_SelectedModule_Name(newName)
        End If
    End Sub

    Friend Sub Change_SelectedModule_Name(ByVal newName As String)
        Me.Cursor = Cursors.AppStarting
        ' --------------------------------------------------------------------- init module
        Dim m As Module_InOut
        m = FindModuleByListLine(SelectedLine)
        If m Is Nothing Then
            Me.Cursor = Cursors.Default
            Return
        End If
        '
        newName = Strings.Left(newName, 16)
        '
        Dim oldName As String = m.GetName
        ' ----------------------------------------------------- try to change the hardware name
        If newName <> oldName Then
            For i As Int32 = 1 To 10
                m.SetName(newName)
                m.WriteNameToHardware()
                ' --------------------------------------------- Wait timeout
                SleepMyThread(100)
                ' --------------------------------------------- Test name
                If m.GetName = newName Then
                    m.SetName(newName)
                    ' ----------------------------------------- if the new name is not in the database
                    If oldName <> "" And FindConfigByName(newName) < 0 Then
                        Dim configid As Int32 = FindConfigByName(oldName)
                        If configid >= 0 Then
                            Dim config As String() = CType(CType(ConfigDatabase(configid), String()).Clone, String())
                            config(0) = config(0).Replace(oldName, newName)
                            AddToConfigDatabase(config)
                            TheSystem_SaveConfigDatabase()
                        End If
                    End If
                    ' ----------------------------------------- 
                    Cmd_Recognize()
                    TheSystem_UpdateListedSubtypes()
                    TheSystem_UpdateConfigDatabase()
                    TheSystem_SaveConfigDatabase()
                    Me.Cursor = Cursors.Default
                    Return
                End If
            Next
            MessageBox.Show("Error writing the name """ + newName + """ to the Master module")
            m.SetName(oldName)
            m.WriteNameToHardware()
            cmb_ModuleNames.Text = oldName
            cmb_ModuleNames.Refresh()
        End If
        Me.Cursor = Cursors.Default
    End Sub

    Private Sub CommSpeed_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txt_CommSpeed.TextChanged
        If Not EventsAreEnabled Then Return
        If SelectedLine < 0 Then Exit Sub
        Timer_10Hz.Stop()
        Dim m As Module_InOut = FindModuleByListLine(SelectedLine)
        If m IsNot Nothing Then
            m.SetSpeed(txt_CommSpeed.NumericValueInteger)
            txt_CommSpeed.Refresh()
            ChangedConfigParams = True
        End If
        Timer_10Hz.Start()
    End Sub

    Private Sub txt_MinChange_TextChanged(ByVal sender As System.Object, ByVal e As System.EventArgs)
        If Not EventsAreEnabled Then Return
        ChangedAppParams = True
    End Sub

    Private Sub txt_MinVariation_Changed(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles txt_MinVariation.TextChanged
        If Not EventsAreEnabled Then Return
        If SelectedLine < 0 Then Exit Sub
        EnableCalibrateButton(TheSystem_CalibrationNeeded())
    End Sub

    Private Sub PinProps_Changed(ByVal sender As System.Object, _
                                 ByVal e As System.EventArgs) _
                                        Handles txt_Slot.TextChanged, _
                                                 txt_MaxValue.TextChanged, _
                                                 txt_MinValue.TextChanged, _
                                                 txt_MaxSpeed.TextChanged, _
                                                 txt_MaxAcc.TextChanged, _
                                                 txt_StepsPerMillim.TextChanged, _
                                                 txt_PwmFrequency.TextChanged, _
                                                 Label_ResponseSpeed.CheckedChanged, _
                                                 txt_ResponseSpeed.TextChanged, _
                                                 txt_ServoMaxTime.TextChanged, _
                                                 txt_ServoMinTime.TextChanged, _
                                                 chk_LogResponse.CheckedChanged, _
                                                 txt_MaxFreq.TextChanged, _
                                                 txt_MinFreq.TextChanged, _
                                                 txt_MinVariation.TextChanged, _
                                                 txt_ProportionalArea.TextChanged
        If Not EventsAreEnabled Then Return
        If SelectedLine < 0 Then Exit Sub
        SetPinProps(False)
        ChangedConfigParams = True
    End Sub

    Private Sub Adc24_PinProps_Changed(ByVal sender As System.Object, _
                                 ByVal e As System.EventArgs) _
                                        Handles chk_Adc24ChBiased.CheckedChanged
        If Not EventsAreEnabled Then Return
        If SelectedLine < 0 Then Exit Sub
        SetPinProps(True)
        ChangedConfigParams = True
    End Sub

    Private Sub SetPinProps(ByVal ConfigAdc24 As Boolean)
        Select Case LineType(SelectedLine)
            Case "Pin", "Adc24", "Gener."
                Dim m As Module_InOut = FindModuleByListLine(SelectedLine)
                Dim p As Pin = FindPinByListLine(SelectedLine)
                Dim n As Int32 = txt_Slot.NumericValueInteger
                Dim str As String
                With MyListView1.Items(SelectedLine)
                    If n <> p.Slot Then
                        p.Slot = n
                        str = p.Slot.ToString()
                        If str <> .SubItems(3).Text Then .SubItems(3).Text = str
                    End If
                    ' TODO - Encoder B
                    If p.GetPinType = Pin.PinTypes.UNUSED OrElse _
                       p.GetPinType = Pin.PinTypes.ENCODER_B OrElse _
                       p.GetPinType = Pin.PinTypes.ENCODER_B_PU OrElse _
                       p.GetPinType = Pin.PinTypes.ADC_24_CH_B Then
                        str = ""
                    Else
                        str = SlotNames(p.Slot)
                    End If
                    If str <> .SubItems(5).Text Then .SubItems(5).Text = str
                End With
                '
                p.Value_Max = CSng(txt_MaxValue.NumericValue)
                p.Value_Min = CSng(txt_MinValue.NumericValue)
                p.AdaptiveSpeed = Label_ResponseSpeed.Checked
                p.ResponseSpeed = txt_ResponseSpeed.NumericValueInteger
                '
                Select Case p.GetPinType
                    Case Pin.PinTypes.SERVO
                        p.MaxTime = CSng(txt_ServoMaxTime.NumericValue)
                        p.MinTime = CSng(txt_ServoMinTime.NumericValue)

                    Case Pin.PinTypes.STEPPER
                        p.MaxSpeed = CSng(txt_MaxSpeed.NumericValue)
                        p.MaxAcc = CSng(txt_MaxAcc.NumericValue)
                        p.StepsPerMillimeter = CSng(txt_StepsPerMillim.NumericValue)
                        ' ------------------------------------------------------------------
                        '  STEPPER PIN CONFIGURATION - By asynchronous communicaton
                        ' ------------------------------------------------------------------
                        p.SendStepperParamsToHardware()
                        ShowStepperErrors(p)
                        ' -----------------------------
                    Case Pin.PinTypes.PWM
                        p.PwmFrequency = CSng(txt_PwmFrequency.NumericValue)
                        InOutModules(p.ModuleId).SendPwmPinFrequency(p.PinIndex, p.PwmFrequency)
                        p.LogResponse = chk_LogResponse.Checked
                        ' ------------------------------------------ ALIGN COUPLED PWM AND SERVO PINS
                        AlignAllCoupledPwmPins(m, p)

                    Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                        p.MinVariation = CSng(txt_MinVariation.NumericValue)
                        p.ProportionalArea = CSng(txt_ProportionalArea.NumericValue)

                    Case Pin.PinTypes.COUNTER, Pin.PinTypes.COUNTER_PU, _
                         Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                        p.MaxFreq = CSng(txt_MaxFreq.NumericValue)
                        p.MinFreq = CSng(txt_MinFreq.NumericValue)
                        p.ConvertToFreq = chk_ConvertToFrequency.Checked

                    Case Pin.PinTypes.ADC_24
                        p.Adc24Sps = CInt(Val(cmb_Adc24Sps.Text))
                        p.Adc24Filter = cmb_Adc24Filter.SelectedIndex

                    Case Pin.PinTypes.ADC_24_CH
                        p.Adc24ChBias = chk_Adc24ChBiased.Checked
                    Case Pin.PinTypes.ADC_24_CH_B
                        p.Adc24ChBias = chk_Adc24ChBiased.Checked

                    Case Pin.PinTypes.ENCODER_B, Pin.PinTypes.ENCODER_B_PU
                        '
                End Select
                '
                If LineType(SelectedLine).StartsWith("Adc24") Then
                    If m.Pins(m.Adc24_MasterPinIndex).GetPinType = Pin.PinTypes.ADC_24 And ConfigAdc24 Then
                        m.Adc24ConfigAndStart()
                    End If
                End If
        End Select
        ' --------------------------------------------------------- show slot conflicts
        TestSlotConflicts()
    End Sub

    ' ====================================================================================================
    '  ALIGN COUPLED PWM AND SERVO
    ' ====================================================================================================
    Private Function CountStepperPins(ByVal m As Module_InOut) As Int32
        For i As Int32 = 0 To m.Pins.Count - 1
            If m.Pins(i).GetPinType = Pin.PinTypes.STEPPER Then
                CountStepperPins += 1
            End If
        Next
    End Function

    Private Function CountServoPins(ByVal m As Module_InOut) As Int32
        For i As Int32 = 0 To m.Pins.Count - 1
            If m.Pins(i).GetPinType = Pin.PinTypes.SERVO Then
                CountServoPins += 1
            End If
        Next
    End Function

    Private Function CountPwmPins(ByVal m As Module_InOut) As Int32
        For i As Int32 = 0 To m.Pins.Count - 1
            If m.Pins(i).GetPinType = Pin.PinTypes.PWM Then
                CountPwmPins += 1
            End If
        Next
    End Function

    Private Function GetPwmPin_Order(ByVal m As Module_InOut, ByVal p As Pin) As Int32
        Dim Order As Int32 = 1
        For i As Int32 = 0 To m.Pins.Count - 1
            If m.Pins(i) Is p Then
                Return Order
            End If
            If m.Pins(i).GetPinType = Pin.PinTypes.SERVO Or m.Pins(i).GetPinType = Pin.PinTypes.PWM Then
                Order += 1
            End If
        Next
    End Function

    Private Function GetPwmPin_FromOrder(ByVal m As Module_InOut, ByVal PwmPinOrder As Int32) As Pin
        Dim Order As Int32 = 1
        For i As Int32 = 0 To m.Pins.Count - 1
            If m.Pins(i).GetPinType = Pin.PinTypes.PWM Then
                If Order = PwmPinOrder Then
                    Return m.Pins(i)
                End If
                Order += 1
            End If
        Next
        Return Nothing
    End Function

    ' ------------------------------------------------------------------------------
    ' Pwm Count            Coupled
    '    9                  8/9 
    '   10                  7/8  9/10 
    '   11                  6/7  8/9  10/11 
    '   12                  5/6  7/8   9/10  11/12 
    '   13                  4/5  6/7   8/9   10/11  12/13
    '   14                  3/4  5/6   7/8    9/10  11/12 13/14
    '   15                  2/3  4/5   6/7    8/9   10/11 12/13  14/15
    '   16                  1/2  3/4   5/6    7/8    9/10 11/12  13/14 15/16 
    ' -------------------------------------------------------------------------------
    '  Stepper are allocated before Servo and Pwm and each Stepper uses two channels
    '  Servo are allocated after Stepper and each one or two Servo uses two channels
    ' -------------------------------------------------------------------------------
    Private Sub AlignAllCoupledPwmPins(ByVal m As Module_InOut, ByVal ThisPin As Pin)
        '
        Dim NumCouplesUsedFromStepperServo As Int32 = CountStepperPins(m) + CountServoPins(m) \ 2
        '
        Dim PwmCount As Int32 = CountPwmPins(m)
        '
        Dim NumCouplesToAlign As Int32 = PwmCount + NumCouplesUsedFromStepperServo - 8
        If NumCouplesToAlign <= 0 Then Return
        If NumCouplesToAlign > 8 Then NumCouplesToAlign = 8
        '
        Dim FirstPinToAlign_Order As Int32 = 8 - NumCouplesUsedFromStepperServo * 2
        '
        Dim ThisPin_Order As Int32 = GetPwmPin_Order(m, ThisPin)
        '
        Dim ReverseCopy As Boolean = ThisPin_Order Mod 2 = PwmCount Mod 2
        'Dim ReverseCopy As Boolean = ThisPin_Order Mod 2 = NumCouplesToAlign Mod 2
        '
        For i As Int32 = FirstPinToAlign_Order To FirstPinToAlign_Order + NumCouplesUsedFromStepperServo * 2 Step 2
            AlignPwmPin(m, i, i + 1, ReverseCopy)
        Next
    End Sub

    Private Sub AlignPwmPin(ByVal m As Module_InOut, _
                            ByVal Pin1_Order As Int32, _
                            ByVal Pin2_Order As Int32, _
                            ByVal ReverseCopy As Boolean)
        Dim PinFrom As Pin
        Dim PinDest As Pin
        If ReverseCopy Then
            PinFrom = GetPwmPin_FromOrder(m, Pin2_Order)
            PinDest = GetPwmPin_FromOrder(m, Pin1_Order)
        Else
            PinFrom = GetPwmPin_FromOrder(m, Pin1_Order)
            PinDest = GetPwmPin_FromOrder(m, Pin2_Order)
        End If
        '
        If PinFrom IsNot Nothing And PinDest IsNot Nothing Then
            If PinDest.PwmFrequency <> PinFrom.PwmFrequency Then
                PinDest.PwmFrequency = PinFrom.PwmFrequency
                InOutModules(PinDest.ModuleId).SendPwmPinFrequency(PinDest.PinIndex, PinDest.PwmFrequency)
            End If
        End If
    End Sub


    ' ====================================================================================================
    '  TEST SLOT CONFLICTS
    ' ====================================================================================================
    Friend Sub TestSlotConflicts()
        If Not ConfigurationIsValid Then Return
        ' ------------------------------------------------------- remove old markers
        For ii As Int32 = 0 To MyListView1.Items.Count - 1
            Dim itm As ListViewItem = MyListView1.Items(ii)
            If itm.SubItems(5).Text = "SLOT CONFLICT" Then
                itm.SubItems(5).Text = SlotNames(FindPinByListLine(ii).Slot)
                ListView_SetLineColors(MyListView1, ii)
            End If
        Next
        ' ------------------------------------------------------- mark new conflicts
        Dim i As Int32 = 0
        Dim dic As Dictionary(Of Int32, Int32) = New Dictionary(Of Int32, Int32)
        For Each m As Module_InOut In InOutModules
            i += 1
            For Each p As Pin In m.Pins
                If p.Direction = Pin.Directions.ModuleToHost Then
                    If dic.ContainsKey(p.Slot) Then
                        MarkSlotConflicts(i)
                        MarkSlotConflicts(dic(p.Slot))
                    End If
                    dic(p.Slot) = i
                End If
                i += 1
            Next
        Next
    End Sub

    Private Sub MarkSlotConflicts(ByVal LineNumber As Int32)
        With MyListView1.Items(LineNumber)
            .SubItems(5).Text = "SLOT CONFLICT"
            If .Selected Then
                .BackColor = Color.FromArgb(49, 106, 197)
                .ForeColor = Color.White
            Else
                .BackColor = Color.FromArgb(255, 150, 150)
                .ForeColor = Color.Black
            End If
        End With
    End Sub

    Friend Sub MarkAllLinesWithDuplicateName(ByVal m As Module_InOut)
        For i As Int32 = 0 To MyListView1.Items.Count - 1
            If FindModuleByListLine(i) Is m Then
                With MyListView1.Items(i)
                    .SubItems(5).Text = "Duplicate name"
                    .BackColor = Color.FromArgb(255, 150, 150)
                    .ForeColor = Color.Black
                End With
            End If
        Next
    End Sub

    ' ====================================================================================================
    '   CONFIG PARAMS LOST FOCUS - SaveConfigDatabase 
    ' ====================================================================================================
    Private ChangedConfigParams As Boolean
    Private Sub ConfigParams_MouseLeave(ByVal sender As System.Object, _
                                        ByVal e As System.EventArgs) _
                                        Handles _
                                        txt_CommSpeed.MouseLeave, _
                                        txt_Slot.MouseLeave, _
                                        txt_MaxValue.MouseLeave, _
                                        txt_MinValue.MouseLeave, _
                                        txt_MaxSpeed.MouseLeave, _
                                        txt_MaxAcc.MouseLeave, _
                                        txt_StepsPerMillim.MouseLeave, _
                                        txt_PwmFrequency.MouseLeave, _
                                        Label_ResponseSpeed.MouseLeave, _
                                        txt_ResponseSpeed.MouseLeave, _
                                        txt_ServoMaxTime.MouseLeave, _
                                        txt_ServoMinTime.MouseLeave, _
                                        chk_LogResponse.MouseLeave, _
                                        txt_MaxFreq.MouseLeave, _
                                        txt_MinFreq.MouseLeave, _
                                        txt_MinVariation.MouseLeave, _
                                        txt_ProportionalArea.MouseLeave, _
                                        chk_Adc24ChBiased.MouseLeave

        If ChangedConfigParams Then
            ChangedConfigParams = False
            TheSystem_UpdateAndSaveConfigDatabase()
        End If
    End Sub

    Private ChangedAppParams As Boolean
    Private Sub txt_MinChange_MouseLeave(ByVal sender As Object, ByVal e As System.EventArgs)
        If ChangedAppParams Then
            ChangedAppParams = False
            Save_INI()
        End If
    End Sub

    Private Sub Special_CheckBoxes_CheckedChanged(ByVal Sender As Object, ByVal e As System.EventArgs) Handles chk_ConvertToFrequency.CheckedChanged
        If Not EventsAreEnabled Then Return
        SetPinProps(False)
        TheSystem_UpdateAndSaveConfigDatabase()
        ShowProps()
    End Sub


    ' ====================================================================================================
    '   COMBO MODULE NAMES
    ' ====================================================================================================
    Private Sub cmb_ModuleNames_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_ModuleNames.DropDown
        Dim oldname As String = cmb_ModuleNames.Items(0).ToString
        cmb_ModuleNames.ItemHeight = 16
        cmb_ModuleNames.Items.Clear()
        Load_ConfigDatabase()
        FillConfigNamesCombo(cmb_ModuleNames)
        Combo_SetIndex_FromString(cmb_ModuleNames, oldname)
    End Sub
    Private Sub cmb_ModuleNames_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_ModuleNames.DropDownClosed
        cmb_ModuleNames.ItemHeight = 12
        Combo_Init(cmb_ModuleNames, cmb_ModuleNames.Text)
        MyListView1.Focus()
    End Sub
    Private Sub cmb_ModuleNames_SelectionChangeCommitted(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_ModuleNames.SelectionChangeCommitted
        Change_SelectedModule_Name(cmb_ModuleNames.Text)
    End Sub


    ' ====================================================================================================
    '   COMBO PIN TYPE
    ' ====================================================================================================
    Friend Sub cmb_PinType_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_PinType.DropDown
        cmb_PinType.ItemHeight = 16
        Dim p As Pin = FindPinByListLine(SelectedLine)
        If p Is Nothing Then Return
        InOutModules(p.ModuleId).FillTypeCombo(cmb_PinType, p)
        Combo_SetIndex_FromString(cmb_PinType, Pin.PinTypeToString(p.GetPinType))
        LimitComboDropDownHeight(cmb_PinType)
    End Sub
    Friend Sub cmb_PinType_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_PinType.DropDownClosed
        cmb_PinType.ItemHeight = 12
    End Sub
    Friend Sub cmb_PinType_SelectionChangeCommitted(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_PinType.SelectionChangeCommitted
        Dim p As Pin = FindPinByListLine(SelectedLine)
        If p Is Nothing Then Return
        Dim m As Module_InOut = InOutModules(p.ModuleId)

        Dim str As String = cmb_PinType.Items(cmb_PinType.SelectedIndex).ToString
        p.SetPinType(Pin.StringToPinType(str))

        ' -------------------------------------------------------------------------- 
        '  STEPPER - Add remove DIR and ENCODER
        ' --------------------------------------------------------------------------
        If p.GetPinType = Pin.PinTypes.STEPPER Then
            Dim p2 As Pin
            p2 = m.Pins(p.PinIndex + 1)
            p2.SetPinType(Pin.PinTypes.STEPPER_DIR)
            SetListViewLineFields(MyListView1.Items(SelectedLine + 1), p2)
            If p.PinIndex < m.Pins.Count - 2 Then
                Dim p3 As Pin
                p3 = m.Pins(p.PinIndex + 2)
                If p3.GetPinType = Pin.PinTypes.ENCODER_B OrElse p3.GetPinType = Pin.PinTypes.ENCODER_B_PU Then
                    p3.SetPinType(Pin.PinTypes.UNUSED)
                End If
                SetListViewLineFields(MyListView1.Items(SelectedLine + 2), p3)
            End If
        End If
        '
        If p.GetPinType <> Pin.PinTypes.STEPPER Then
            If p.PinIndex < m.Pins.Count - 1 Then
                Dim p2 As Pin
                p2 = m.Pins(p.PinIndex + 1)
                If p2.GetPinType = Pin.PinTypes.STEPPER_DIR Then
                    p2.SetPinType(Pin.PinTypes.UNUSED)
                End If
                If SelectedLine + 1 < MyListView1.Items.Count Then
                    SetListViewLineFields(MyListView1.Items(SelectedLine + 1), p2)
                End If
            End If
        End If

        ' -------------------------------------------------------------------------- 
        '  ENCODER - Add remove ENCODER_B
        ' --------------------------------------------------------------------------
        If p.GetPinType = Pin.PinTypes.ENCODER_A Or _
           p.GetPinType = Pin.PinTypes.ENCODER_A_PU Then
            Dim p2 As Pin
            p2 = m.Pins(p.PinIndex + 1)
            If p.GetPinType = Pin.PinTypes.ENCODER_A Then
                p2.SetPinType(Pin.PinTypes.ENCODER_B)
            Else
                p2.SetPinType(Pin.PinTypes.ENCODER_B_PU)
            End If
            If SelectedLine + 1 < MyListView1.Items.Count Then
                SetListViewLineFields(MyListView1.Items(SelectedLine + 1), p2)
            End If
        End If
        '
        If p.GetPinType <> Pin.PinTypes.ENCODER_A And _
           p.GetPinType <> Pin.PinTypes.ENCODER_A_PU Then
            If p.PinIndex < m.Pins.Count - 1 Then
                Dim p2 As Pin
                p2 = m.Pins(p.PinIndex + 1)
                If p2.GetPinType = Pin.PinTypes.ENCODER_B Or _
                   p2.GetPinType = Pin.PinTypes.ENCODER_B_PU Then
                    p2.SetPinType(Pin.PinTypes.UNUSED)
                End If
                If SelectedLine + 1 < MyListView1.Items.Count Then
                    SetListViewLineFields(MyListView1.Items(SelectedLine + 1), p2)
                End If
            End If
        End If

        Adc24_AlignCoupledPinParams(p)
        Adc24_TestSteppers(m)


        ' -------------------------------------------------------------------------- 
        '  FINAL UPDATES
        ' --------------------------------------------------------------------------
        InOutModules(p.ModuleId).SetupModulePins()
        ' -------------------------------------------------------------------------- 
        Combo_Init(cmb_PinType, Pin.PinTypeToString(p.GetPinType))
        ' -------------------------------------------------------------------------- set PinType and Direction
        TheSystem_UpdateListedSubtypes()
        SetListViewLineFields(MyListView1.Items(SelectedLine), p)
        ' -------------------------------------------------------------------------- set Color
        ListView_SetAllLineColors()
        ' -------------------------------------------------------------------------- set pin props
        ShowProps()
        SetPinProps(True)
        TheSystem_UpdateAndSaveConfigDatabase()
        ' -------------------------------------------------------------------------- 
        EnableCalibrateButton(TheSystem_CalibrationNeeded())
        If btn_Calibrate.Enabled Then
            Select Case p.GetPinType
                Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                    DoCalibration()
            End Select
        End If
    End Sub

    Private Sub SetListViewLineFields(ByVal l As ListViewItem, ByVal p As Pin)
        If SelectedLine < 0 Then Return
        With l
            .SubItems(2).Text = Pin.PinTypeToString(p.GetPinType)
            .SubItems(3).Text = If(p.Direction = Pin.Directions.Unused, "", p.Slot.ToString())
            .SubItems(4).Text = ""
            ' 
            Dim str As String
            If p.GetPinType = Pin.PinTypes.UNUSED Then
                str = ""
            Else
                str = SlotNames(p.Slot)
            End If
            If str <> .SubItems(5).Text Then .SubItems(5).Text = str
        End With
    End Sub

    Private Sub Adc24_TestSteppers(ByVal m As Module_InOut)
        'If m.Pins(Adc24_MasterPinIndex).GetPinType = Pin.PinTypes.ADC_24 Then
        '    For Each p As Pin In m.Pins
        '        If p.GetPinType = Pin.PinTypes.STEPPER Then
        '            MessageBox.Show(Msg_Warning + vbCrLf + vbCrLf + _
        '                            Msg_StepperWithAdc24, _
        '                            Msg_HalMessage, _
        '                            MessageBoxButtons.OK, _
        '                            MessageBoxIcon.Warning)
        '            Return
        '        End If
        '    Next
        'End If
    End Sub


    ' ====================================================================================================
    '   COMBO Adc24Sps
    ' ====================================================================================================
    Private Sub cmb_Adc24Sps_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24Sps.DropDown
        cmb_Adc24Sps.ItemHeight = 16
    End Sub
    Private Sub cmb_Adc24Sps_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24Sps.DropDownClosed
        cmb_Adc24Sps.ItemHeight = 12
    End Sub
    Friend Sub cmb_Adc24Sps_SelectionChangeCommitted(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24Sps.SelectionChangeCommitted
        Dim p As Pin = FindPinByListLine(SelectedLine)
        If p Is Nothing Then Return
        p.Adc24Sps = CInt(Val(cmb_Adc24Sps.Text))
        SetPinProps(True)
        TheSystem_UpdateAndSaveConfigDatabase()
    End Sub

    ' ====================================================================================================
    '   COMBO Adc24Filter
    ' ====================================================================================================
    Private Sub cmb_Adc24Filter_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24Filter.DropDown
        cmb_Adc24Filter.ItemHeight = 16
    End Sub
    Private Sub cmb_Adc24Filter_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24Filter.DropDownClosed
        cmb_Adc24Filter.ItemHeight = 12
    End Sub
    Friend Sub cmb_Adc24Filter_SelectionChangeCommitted(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24Filter.SelectionChangeCommitted
        Dim p As Pin = FindPinByListLine(SelectedLine)
        If p Is Nothing Then Return
        p.Adc24Filter = cmb_Adc24Filter.SelectedIndex
        SetPinProps(True)
        TheSystem_UpdateAndSaveConfigDatabase()
    End Sub

    ' ====================================================================================================
    '   COMBO Adc24ChType
    ' ====================================================================================================
    Private Sub cmb_Adc24ChType_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24ChType.DropDown
        cmb_Adc24ChType.ItemHeight = 16
    End Sub
    Private Sub cmb_Adc24ChType_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24ChType.DropDownClosed
        cmb_Adc24ChType.ItemHeight = 12
    End Sub
    Friend Sub cmb_Adc24ChType_SelectionChangeCommitted(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24ChType.SelectionChangeCommitted
        Dim p As Pin = FindPinByListLine(SelectedLine)
        If p Is Nothing Then Return
        p.Adc24ChType = cmb_Adc24ChType.SelectedIndex
        Adc24_AlignCoupledPinParams(p)
        InOutModules(p.ModuleId).SetupModulePins()
        ShowProps()
        SetPinProps(True)
        TheSystem_UpdateAndSaveConfigDatabase()
    End Sub

    ' ====================================================================================================
    '   COMBO Adc24ChGain
    ' ====================================================================================================
    Private Sub cmb_Adc24ChGain_DropDown(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24ChGain.DropDown
        cmb_Adc24ChGain.ItemHeight = 16
    End Sub
    Private Sub cmb_Adc24ChGain_DropDownClosed(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24ChGain.DropDownClosed
        cmb_Adc24ChGain.ItemHeight = 12
    End Sub
    Friend Sub cmb_Adc24ChGain_SelectionChangeCommitted(ByVal sender As Object, ByVal e As System.EventArgs) Handles cmb_Adc24ChGain.SelectionChangeCommitted
        Dim p As Pin = FindPinByListLine(SelectedLine)
        If p Is Nothing Then Return
        p.Adc24ChGain = cmb_Adc24ChGain.SelectedIndex
        Adc24_AlignCoupledPinParams(p)
        SetPinProps(True)
        TheSystem_UpdateAndSaveConfigDatabase()
    End Sub

    ' ====================================================================================================
    '   Adc24_AlignCoupledPinParams
    ' ====================================================================================================
    Private Sub Adc24_AlignCoupledPinParams(ByVal p As Pin)

        Dim m As Module_InOut = InOutModules(p.ModuleId)

        ' ----------------------------------------------------- return if not adc24 pin
        If p.PinIndex < m.Adc24_MasterPinIndex + 1 Then Return
        If p.PinIndex > m.Adc24_MasterPinIndex + 19 Then Return
        ' ----------------------------------------------------- return if not coupled 
        If (p.PinIndex - m.Adc24_FirstChannelIndex) Mod 2 = 0 Then
            If p.PinIndex >= m.Pins.Count - 1 Then Return
        End If
        ' ----------------------------------------------------- set CH and CH_B to coupled pin
        If (p.PinIndex - m.Adc24_FirstChannelIndex) Mod 2 = 0 Then
            Dim p2 As Pin = m.Pins(p.PinIndex + 1)
            If p.Adc24ChType = 0 Then ' if differential
                If p2.GetPinType = Pin.PinTypes.ADC_24_CH Then
                    p2.SetPinType(Pin.PinTypes.ADC_24_CH_B)
                    SetListViewLineFields(MyListView1.Items(SelectedLine + 1), p2)
                End If
            Else
                If p2.GetPinType = Pin.PinTypes.ADC_24_CH_B Then
                    p2.SetPinType(Pin.PinTypes.ADC_24_CH)
                    SetListViewLineFields(MyListView1.Items(SelectedLine + 1), p2)
                End If
            End If
        End If
        ' ----------------------------------------------------- set CH and CH_B to selected pin
        If (p.PinIndex - m.Adc24_FirstChannelIndex) Mod 2 = 1 Then
            If p.Adc24ChType = 0 Then ' if differential
                If p.GetPinType = Pin.PinTypes.ADC_24_CH Then
                    m.Pins(p.PinIndex).SetPinType(Pin.PinTypes.ADC_24_CH_B)
                    SetListViewLineFields(MyListView1.Items(SelectedLine), p)
                End If
            Else
                If p.GetPinType = Pin.PinTypes.ADC_24_CH_B Then
                    ' ------------------------------------- save and restore old values
                    Dim oldChType As Int32 = p.Adc24ChType
                    Dim oldChGain As Int32 = p.Adc24ChGain
                    m.Pins(p.PinIndex).SetPinType(Pin.PinTypes.ADC_24_CH)
                    p.Adc24ChType = oldChType
                    p.Adc24ChGain = oldChGain
                    SetListViewLineFields(MyListView1.Items(SelectedLine), p)
                End If
            End If
        End If
        ' ----------------------------------------------------- copy to coupled pin
        If (p.PinIndex - m.Adc24_FirstChannelIndex) Mod 2 = 0 Then
            If p.PinIndex + 1 < m.Pins.Count Then
                m.Pins(p.PinIndex + 1).Adc24ChType = p.Adc24ChType
                m.Pins(p.PinIndex + 1).Adc24ChGain = p.Adc24ChGain
            End If
        Else
            m.Pins(p.PinIndex - 1).Adc24ChType = p.Adc24ChType
            m.Pins(p.PinIndex - 1).Adc24ChGain = p.Adc24ChGain
        End If

    End Sub


    ' ===========================================================================================================
    '   SHOW PROPS
    ' ===========================================================================================================
    Friend Sub ListView_SetAllLineColors()
        ' -------------------------------------------------------------------- select the listview "SelectedLine"
        If SelectedLine >= 0 And SelectedLine < MyListView1.Items.Count Then
            MyListView1.Items(SelectedLine).Selected = True
        End If
        ' -------------------------------------------------------------------- show listview colors
        For i As Int32 = 0 To MyListView1.Items.Count - 1
            ListView_SetLineColors(MyListView1, i)
        Next
    End Sub

    Private Sub ListView_SetLineColors(ByVal lv As ListView, ByVal line As Int32)
        If line < 0 Or line >= lv.Items.Count Then Return
        '
        Dim fc As Color
        Dim bc As Color
        If lv.Items(line).Selected Then
            ' ----------------------------------------------------------------- selected colors
            bc = Color.FromArgb(49, 106, 197)
            fc = Color.White
        Else
            ' ----------------------------------------------------------------- deselected colors
            If line >= 0 Then
                Select Case LineType(line)
                    Case "Module"
                        ' TODO colors
                        TheSystem_UpdateListColors(line)
                        Return
                    Case "Pin"
                        If LineSubType(line) <> "Unused" Then
                            If FindPinByListLine(line).Direction = Pin.Directions.HostToModule Then
                                bc = Color.FromArgb(245, 245, 215) '(255, 255, 225)
                            Else
                                bc = Color.FromArgb(225, 245, 230) '(235, 255, 240)
                            End If
                        Else
                            bc = Color.FromArgb(255, 255, 255)
                        End If
                    Case "Adc24"
                        If LineSubType(line) <> "Unused" Then
                            bc = Color.FromArgb(200, 235, 180)
                        Else
                            bc = Color.FromArgb(255, 255, 255)
                        End If
                    Case "Gener."
                        If LineSubType(line) <> "Unused" Then
                            If FindPinByListLine(line).Direction = Pin.Directions.HostToModule Then
                                bc = Color.FromArgb(235, 230, 140)
                            Else
                                bc = Color.FromArgb(205, 240, 140)
                            End If
                        Else
                            bc = Color.FromArgb(255, 255, 255)
                        End If
                End Select
                fc = Color.Black
            End If
        End If
        lv.Items(line).ForeColor = fc
        lv.Items(line).BackColor = bc
    End Sub

    Private Function LineType(ByVal line As Int32) As String
        If line < 0 Then Return ""
        If line > MyListView1.Items.Count - 1 Then Return ""
        Return MyListView1.Items(line).SubItems(0).Text.Trim
    End Function
    Private Function LineSubType(ByVal line As Int32) As String
        Return MyListView1.Items(line).SubItems(2).Text.Trim
    End Function

    Friend SelectedLine As Int32 = -1
    Private Sub MyListView1_SelectedIndexChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles MyListView1.SelectedIndexChanged
        If Not EventsAreEnabled Then Return
        LockFormUpdate(Me)
        SyncLock EventsLock
            ' --------------------------------------------------------- change line colors and update "SelectedLine"
            If MyListView1.SelectedIndices.Count > 0 Then
                SelectedLine = MyListView1.SelectedIndices(0)
                If ConfigurationIsValid Then ListView_SetLineColors(MyListView1, SelectedLine)
            Else
                If ConfigurationIsValid Then ListView_SetLineColors(MyListView1, SelectedLine)
                SelectedLine = -1
            End If
            ' --------------------------------------------------------- show line props
            ShowProps()
            ' --------------------------------------------------------- show slot conflicts
            TestSlotConflicts()
        End SyncLock
        UnlockFormUpdate()
        If SelectedLine >= 0 Then Refresh()
    End Sub

    Private Sub ShowProps()
        '
        EventsAreEnabled = False
        '
        GroupBox_PinProps.Visible = False
        GroupBox_TouchProps.Visible = False
        GroupBox_ServoPwmProps.Visible = False
        GroupBox_StepperProps.Visible = False
        GroupBox_PwmProps.Visible = False
        GroupBox_FrequencyProps.Visible = False
        GroupBox_Adc24Props.Visible = False
        GroupBox_Adc24ChProps.Visible = False
        '
        If SelectedLine < 0 Or SelectedLine > MyListView1.Items.Count - 1 Then
            GroupBox_ModuleProps.Enabled = False
            txt_CommSpeed.Text = "-"
        Else
            GroupBox_ModuleProps.Enabled = True
            '
            Dim m As Module_InOut = FindModuleByListLine(SelectedLine)
            Dim p As Pin = FindPinByListLine(SelectedLine)
            '
            If Not ConfigurationIsValid Then
                GroupBox_ModuleProps.Height = cmb_ModuleNames.Bottom + 6
                If m IsNot Nothing Then
                    Combo_Init(cmb_ModuleNames, m.GetName)
                Else
                    Combo_Init(cmb_ModuleNames, "")
                    GroupBox_ModuleProps.Enabled = False
                    txt_CommSpeed.Text = "-"
                End If
            Else
                GroupBox_ModuleProps.Height = Label_CommSpeed.Bottom + 6 '130
                '
                Select Case LineType(SelectedLine)
                    Case "Module"
                        GroupBox_PinProps.Visible = False
                        GroupBox_ServoPwmProps.Visible = False
                    Case "Pin", "Adc24", "Gener."
                        GroupBox_PinProps.Visible = True
                        GroupBox_PinProps.Left = GroupBox_ModuleProps.Left
                        GroupBox_PinProps.Height = txt_ResponseSpeed.Bottom + 8
                End Select
                '
                If m IsNot Nothing Then
                    Combo_Init(cmb_ModuleNames, m.GetName)
                    txt_CommSpeed.NumericValueInteger = m.CommSpeed
                    '
                    ToolTips_Init()
                    '
                    If p IsNot Nothing Then
                        Label_MaxValue.Text = Msg_MaxValue
                        Label_MinValue.Text = Msg_MinValue
                        Combo_Init(cmb_PinType, Pin.PinTypeToString(p.GetPinType))
                        txt_Slot.NumericValue = p.Slot
                        txt_MaxValue.NumericValue = p.Value_Max
                        txt_MinValue.NumericValue = p.Value_Min
                        Label_ResponseSpeed.Checked = p.AdaptiveSpeed
                        txt_ResponseSpeed.NumericValueInteger = p.ResponseSpeed
                        txt_MaxValue.Visible = True
                        Select Case p.GetPinType
                            Case Pin.PinTypes.UNUSED
                                GroupBox_PinProps.Height = cmb_PinType.Bottom + 8
                                ' ------------------------------------------------------------------ OUT
                            Case Pin.PinTypes.SERVO
                                GroupBox_ServoPwmProps.Text = Msg_ServoProps
                                txt_ServoMaxTime.MaxValue = 4000
                                txt_ServoMaxTime.NumericValue = p.MaxTime
                                txt_ServoMinTime.NumericValue = p.MinTime
                                GroupBox_ServoPwmProps.Left = GroupBox_PinProps.Left
                                GroupBox_ServoPwmProps.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_ServoPwmProps.Height = txt_ServoMinTime.Bottom + 8
                                GroupBox_ServoPwmProps.Visible = True
                            Case Pin.PinTypes.STEPPER
                                Label_MaxValue.Text = Msg_MaxValueStepper
                                Label_MinValue.Text = Msg_MinValueStepper
                                txt_MaxSpeed.NumericValue = p.MaxSpeed
                                txt_MaxAcc.NumericValue = p.MaxAcc
                                txt_StepsPerMillim.NumericValue = p.StepsPerMillimeter
                                GroupBox_StepperProps.Left = GroupBox_PinProps.Left
                                GroupBox_StepperProps.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_StepperProps.Height = Label_StepsPerMillim.Bottom + 8
                                GroupBox_StepperProps.Visible = True
                                ShowStepperErrors(p)
                            Case Pin.PinTypes.PWM
                                txt_PwmFrequency.NumericValue = p.PwmFrequency
                                chk_LogResponse.Checked = p.LogResponse
                                GroupBox_PwmProps.Left = GroupBox_PinProps.Left
                                GroupBox_PwmProps.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_PwmProps.Height = chk_LogResponse.Bottom + 7
                                GroupBox_PwmProps.Visible = True
                                ' ------------------------------------------------------------------ IN
                            Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                                txt_MinVariation.NumericValue = p.MinVariation
                                txt_ProportionalArea.NumericValue = p.ProportionalArea
                                GroupBox_TouchProps.Left = GroupBox_PinProps.Left
                                GroupBox_TouchProps.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_TouchProps.Visible = True
                            Case Pin.PinTypes.COUNTER, Pin.PinTypes.COUNTER_PU, _
                                 Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                                chk_ConvertToFrequency.Checked = p.ConvertToFreq
                                txt_MaxFreq.NumericValue = p.MaxFreq
                                txt_MinFreq.NumericValue = p.MinFreq
                                If chk_ConvertToFrequency.Checked Then
                                    GroupBox_PinProps.Height = txt_ResponseSpeed.Bottom + 10
                                    GroupBox_FrequencyProps.Height = txt_MinFreq.Bottom + 8
                                Else
                                    GroupBox_PinProps.Height = txt_Slot.Bottom + 7
                                    GroupBox_FrequencyProps.Height = chk_ConvertToFrequency.Bottom + 10
                                End If
                                GroupBox_FrequencyProps.Left = GroupBox_PinProps.Left
                                GroupBox_FrequencyProps.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_FrequencyProps.Visible = True
                            Case Pin.PinTypes.STEPPER_DIR
                                GroupBox_PinProps.Height = txt_Slot.Bottom + 7
                                txt_MaxValue.Visible = False
                            Case Pin.PinTypes.ADC_24
                                GroupBox_PinProps.Height = cmb_PinType.Bottom + 8
                                Combo_SetIndex_FromString(cmb_Adc24Sps, p.Adc24Sps.ToString)
                                If cmb_Adc24Sps.Text = "" Then cmb_Adc24Sps.SelectedIndex = 0
                                Combo_SetIndex(cmb_Adc24Filter, p.Adc24Filter)
                                GroupBox_Adc24Props.Left = GroupBox_PinProps.Left
                                GroupBox_Adc24Props.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_Adc24Props.Height = cmb_Adc24Filter.Bottom + 7
                                GroupBox_Adc24Props.Visible = True
                            Case Pin.PinTypes.ADC_24_CH
                                Combo_SetIndex(cmb_Adc24ChType, p.Adc24ChType)
                                Combo_SetIndex(cmb_Adc24ChGain, p.Adc24ChGain)
                                chk_Adc24ChBiased.Checked = p.Adc24ChBias
                                GroupBox_Adc24ChProps.Left = GroupBox_PinProps.Left
                                GroupBox_Adc24ChProps.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_Adc24ChProps.Height = chk_Adc24ChBiased.Bottom + 7
                                GroupBox_Adc24ChProps.Visible = True
                            Case Pin.PinTypes.ADC_24_CH_B
                                GroupBox_PinProps.Height = cmb_PinType.Bottom + 8
                                Combo_SetIndex(cmb_Adc24ChType, p.Adc24ChType)
                                Combo_SetIndex(cmb_Adc24ChGain, p.Adc24ChGain)
                                chk_Adc24ChBiased.Checked = p.Adc24ChBias
                                GroupBox_Adc24ChProps.Left = GroupBox_PinProps.Left
                                GroupBox_Adc24ChProps.Top = GroupBox_PinProps.Bottom + 6
                                GroupBox_Adc24ChProps.Height = chk_Adc24ChBiased.Bottom + 7
                                GroupBox_Adc24ChProps.Visible = True
                            Case Pin.PinTypes.ENCODER_A, Pin.PinTypes.ENCODER_A_PU
                                GroupBox_PinProps.Height = txt_Slot.Bottom + 7
                                txt_MaxValue.Visible = False
                            Case Pin.PinTypes.ENCODER_B, Pin.PinTypes.ENCODER_B_PU
                                GroupBox_PinProps.Height = cmb_PinType.Bottom + 8
                        End Select
                    End If
                End If
            End If
        End If
        '
        EventsAreEnabled = True
    End Sub

    Friend Sub ShowStepperErrors(ByVal p As Pin)
        If p.StepperPinDataOK Then
            Label_MaxSpeed.BackColor = Label_MaxAcc.BackColor
            Label_StepsPerMillim.BackColor = Label_MaxAcc.BackColor
            Label_MaxSpeed.ForeColor = Label_MaxAcc.ForeColor
            Label_StepsPerMillim.ForeColor = Label_MaxAcc.ForeColor
        Else
            Label_MaxSpeed.BackColor = Color.Yellow
            Label_StepsPerMillim.BackColor = Color.Yellow
            Label_MaxSpeed.ForeColor = Color.Red
            Label_StepsPerMillim.ForeColor = Color.Red
        End If
    End Sub


    ' ==============================================================================================================
    '   ToolTips
    ' ==============================================================================================================
    Private WithEvents ttp1 As ToolTip = New ToolTip
    Private Sub ToolTips_Init()
        ttp1.ShowAlways = True
        ttp1.UseAnimation = True
        ttp1.UseFading = True
        ttp1.IsBalloon = True
        ttp1.AutoPopDelay = 32700
        ttp1.InitialDelay = 500
        ttp1.ReshowDelay = 500
        ' --------------------------------------------------------------- Add here Controls and ToolTip-Text
        Dim m As Module_InOut = FindModuleByListLine(SelectedLine)
        If m IsNot Nothing Then
            ttp1.SetToolTip(txt_CommSpeed, _
                 Msg_CommSpeed1 & vbCrLf & _
                 "- - - - - - - - - - - - - - -" & vbCrLf & _
                 "  1  =  10 fps" & vbCrLf & _
                 "  2  =  20 fps" & vbCrLf & _
                 "  3  =  30 fps" & vbCrLf & _
                 "  4  =  50 fps" & vbCrLf & _
                 "  5  =  60 fps" & vbCrLf & _
                 "  6  = 100 fps" & vbCrLf & _
                 "  7  = 150 fps" & vbCrLf & _
                 "  8  = 200 fps" & vbCrLf & _
                 "  9  = 300 fps" & vbCrLf & _
                 "10  =  400 fps" & vbCrLf & _
                 "11  =  500 fps" & vbCrLf & _
                 "12  =  Max fps")
        End If
        ttp1.SetToolTip(Label_ResponseSpeed, Msg_Filter1 & vbCrLf & _
                                        Msg_Filter2 & vbCrLf & _
                                        Msg_Filter3 & vbCrLf & _
                                        Msg_Filter4 & vbCrLf & _
                                        Msg_Filter5)
        ToolStripButton_Recognize.ToolTipText = Msg_Recognize1
        ToolStripButton_Validate.ToolTipText = Msg_Validate1
        ToolStripButton_BeepOnErrors.ToolTipText = Msg_BeepOnErrors1
        ToolStripButton_CommOptions.ToolTipText = Msg_CommOptions1
    End Sub

    ' --------------------- this corrects the XP bug that causes tooltip disappearing for ever 
    Private Sub controls_MouseEnter(ByVal sender As Object, _
                                   ByVal e As System.EventArgs) Handles txt_CommSpeed.MouseEnter
        ttp1.Active = False
        ttp1.Active = True
    End Sub


    ' ==============================================================================================================
    '   AUTO CALIBRATION
    ' ==============================================================================================================
    Friend Sub DoCalibration()
        btn_Calibrate.Checked = True
        btn_Calibrate.Refresh()
        TheSystem_Calibrate()
        System.Threading.Thread.Sleep(100)
        btn_Calibrate.Checked = False
    End Sub

    Friend Sub EnableCalibrateButton(ByVal enable As Boolean)
        btn_Calibrate.Enabled = enable
    End Sub

    Private Sub btn_Calibrate_ClickButtonArea(ByVal Sender As System.Object, ByVal e As System.EventArgs) Handles btn_Calibrate.ClickButtonArea
        If Not EventsAreEnabled Then Return
        EventsAreEnabled = False
        DoCalibration()
        btn_Calibrate.Refresh()
        EventsAreEnabled = True
    End Sub

End Class


